/*****************************************************************************
*
* Atmel Corporation
*
* File              : USI_TWI_Master.c
* Compiler          : AVRGCC Toolchain version 3.4.2
* Revision          : $Revision: 992 $
* Date              : $Date: 2013-11-07 $
* Updated by        : $Author: Atmel $
*
* Support mail      : avr@atmel.com
*
* Supported devices : All device with USI module can be used.
*                     The example is written for the ATmega169, ATtiny26 and ATtiny2313
*
* AppNote           : AVR310 - Using the USI module as a TWI Master
*
* Description       : This is an implementation of an TWI master using
*                     the USI module as basis. The implementation assumes the AVR to
*                     be the only TWI master in the system and can therefore not be
*                     used in a multi-master system.
* Usage             : Initialize the USI module by calling the USI_TWI_Master_Initialise() 
*                     function. Hence messages/data are transceived on the bus using
*                     the USI_TWI_Transceive() function. The transceive function 
*                     returns a status byte, which can be used to evaluate the 
*                     success of the transmission.
*
****************************************************************************/
//	2017/06	@jsdiy
//・I2C仕様書とATmega88Aのデータシート掲載のタイミングチャートを参考にAVR310を改造。
//・SDA使用後はPORT_USIをHi、USIDRを0xFFにするのが安定的に正常動作させるポイント。
//****************************************************************************

#include <avr/io.h>
#include <util/delay.h>
#include "usiTwiMaster.h"

#ifdef TWI_STANDARD_MODE
	#define T_LOW			5.7	// >4.7us	Low Period of the SCL Clock
	#define T_HIGH			5.0	// >4.0us	High period of the SCL clock
	#define T_HOLD_START	5.0	// >4.0us	Hold Time (repeated) START Condition
	#define T_SETUP_START	5.7	// >4.7us	Set-up time for a repeated START condition
	#define T_SETUP_STOP	5.0	// >4.0us	Setup time for STOP condition
	#define T_BUS_FREE		5.7	// >4.7us	Bus free time between a STOP and START condition
#else	//TWI_FAST_MODE
	#define T_LOW			2.3	// >1.3us	Low Period of the SCL Clock
	#define T_HIGH			1.6	// >0.6us	High period of the SCL clock
	#define T_HOLD_START	1.6	// >0.6us	Hold Time (repeated) START Condition
	#define T_SETUP_START	1.6	// >0.6us	Set-up time for a repeated START condition
	#define T_SETUP_STOP	1.6	// >0.6us	Setup time for STOP condition
	#define T_BUS_FREE		2.3	// >1.3us	Bus free time between a STOP and START condition
#endif

// Defines controling code generating
//#define PARAM_VERIFICATION
//#define NOISE_TESTING
//#define SIGNAL_VERIFY

#define SLA_W_BIT	0
#define SLA_R_BIT	1

//エラーコード
enum
{
	USI_TWI_SUCCESS = 0,
	USI_TWI_NO_DATA,			// Transmission buffer is empty
	USI_TWI_DATA_OUT_OF_BOUND,	// Transmission buffer is outside SRAM space
	USI_TWI_UE_START_CON,		// Unexpected Start Condition
	USI_TWI_UE_STOP_CON,		// Unexpected Stop Condition
	USI_TWI_UE_DATA_COL,		// Unexpected Data Collision (arbitration)
	USI_TWI_NO_ACK_ON_DATA,		// The slave did not acknowledge  all data
	USI_TWI_NO_ACK_ON_ADDRESS,	// The slave did not acknowledge  the address
	USI_TWI_MISSING_START_CON,	// Generated Start Condition not detected on bus
	USI_TWI_MISSING_STOP_CON	// Generated Stop Condition not detected on bus
};

//エラー状態
static	uint8_t	errorState = USI_TWI_SUCCESS;

//プロトタイプ宣言
static	void	MasterTransmit(uint8_t bvBitShift);
static	uint8_t	MasterReceive(uint8_t bvBitShift);


//USI-TWIの初期化
//・バスがフリーの状態＝SCL/SDAどちらもHi.(I2Cの仕様)
void	TWI_MasterInit()
{
	PORT_USI |= (1<<PIN_USI_SDA);	// Enable pullup on SDA, to set high as released state.
	PORT_USI |= (1<<PIN_USI_SCL);	// Enable pullup on SCL, to set high as released state.

	DDR_USI |= (1<<PIN_USI_SCL);	// Enable SCL as output.
	DDR_USI |= (1<<PIN_USI_SDA);	// Enable SDA as output.

	USIDR = 0xFF;						// Preload dataregister with "released level" data.

	USICR = (0<<USISIE)|(0<<USIOIE)|	// Disable Interrupts.
			(1<<USIWM1)|(0<<USIWM0)|	// Set USI in Two-wire mode.
			(1<<USICS1)|(0<<USICS0)|(1<<USICLK)|	// Software strobe as counter clock source
			(0<<USITC);

	USISR = (1<<USISIF)|(1<<USIOIF)|(1<<USIPF)|(1<<USIDC)|	// Clear flags,
			(0x00<<USICNT0);								// and reset counter.
}

//スタートを送信する
//・SCLがHiの状態でSDAをHi→Loにする。
//・スタート再送時(データ送信続行中のためストップをまだ送信していない状態)を想定し、まずSCLをHiにする。
//戻り値	非0:成功, 0:失敗
uint8_t	TWI_Start()
{
	errorState = USI_TWI_SUCCESS;

	/* Release SCL to ensure that (repeated) Start can be performed */
	PORT_USI |= (1<<PIN_USI_SCL);	// Release SCL.
	_delay_us(T_SETUP_START);

	/* Generate Start Condition */
	PORT_USI &= ~(1<<PIN_USI_SDA);	// Force SDA LOW.
	_delay_us(T_HOLD_START);
	PORT_USI &= ~(1<<PIN_USI_SCL);	// Pull SCL LOW.
	_delay_us(T_LOW);

	PORT_USI |= (1<<PIN_USI_SDA);	// Release SDA.	(必要)

#ifdef SIGNAL_VERIFY
	if( !(USISR & (1<<USISIF)) ) { errorState = USI_TWI_MISSING_START_CON; }
#endif

	return (errorState == USI_TWI_SUCCESS);
}

//Prepare register value to: Clear flags, and set USI to shift 8 bits i.e. count 16 clock edges.
//最後の項、4ビットカウンタを0x00(0)で初期化することで残り16クロック上下端（8ビット分)動作する
#define BV_USISR_8BITSHIFT	( (1<<USISIF)|(1<<USIOIF)|(1<<USIPF)|(1<<USIDC)|(0x00<<USICNT0) )

//Prepare register value to: Clear flags, and set USI to shift 1 bit i.e. count 2 clock edges.
//最後の項、4ビットカウンタを0x0E(14)で初期化することで残り2クロック上下端（1ビット分)動作する
#define BV_USISR_1BITSHIFT	( (1<<USISIF)|(1<<USIOIF)|(1<<USIPF)|(1<<USIDC)|(0x0E<<USICNT0) )

//Prepare clocking.
#define BV_USICR_SETCLOCK	\
  ( (0<<USISIE)|(0<<USIOIE)|				/* Interrupts disabled */	\
	(1<<USIWM1)|(0<<USIWM0)|				/* Set USI in Two-wire mode. */	\
	(1<<USICS1)|(0<<USICS0)|(1<<USICLK)|	/* Software clock strobe as source. */	\
	(1<<USITC) )							/* Toggle Clock Port. */

//スレーブデバイスのアドレス／任意のデータを送信する
//・スレーブ側がSCLをストレッチする場合があるのでSCLの変化を監視する必要がある。
static	void	MasterTransmit(uint8_t bvBitShift)
{
	USISR = bvBitShift;		//何ビットシフト送出させるか
	do
	{
		USICR = BV_USICR_SETCLOCK;				// Generate positive SCL edge.
		while( !(PIN_USI & (1<<PIN_USI_SCL)) );	// Wait for SCL to go high.
		_delay_us(T_HIGH);
		USICR = BV_USICR_SETCLOCK;				// Generate negative SCL edge.
		while( (PIN_USI & (1<<PIN_USI_SCL)) );	// Wait for SCL to go low.
		_delay_us(T_LOW);
	}
	while( !(USISR & (1<<USIOIF)) );	// Check for transfer complete.
}

//任意のデータを受信する
static	uint8_t	MasterReceive(uint8_t bvBitShift)
{
	uint8_t	data;

	DDR_USI &= ~(1<<PIN_USI_SDA);	// Enable SDA as input.
	MasterTransmit(bvBitShift);
	data = USIDR;					// Read out data.
	USIDR = 0xFF;					// Release SDA.	(必要)
	DDR_USI |= (1<<PIN_USI_SDA);	// Enable SDA as output.

	return data;	// Return the data from the USIDR
}

//スレーブデバイスのアドレスを指定する：書き込み方向
//引数:	スレーブアドレス(7bit)
//戻り値	非0:成功, 0:失敗
uint8_t	TWI_SlaveAddrW(uint8_t sla)
{
	uint8_t	isOK = TWI_WriteByte((sla << 1) | SLA_W_BIT);
	if (!isOK) { errorState = USI_TWI_NO_ACK_ON_ADDRESS; }
	return isOK;
}

//スレーブデバイスのアドレスを指定する：読み込み方向
//引数:	スレーブアドレス(7bit)
//戻り値	非0:成功, 0:失敗
uint8_t	TWI_SlaveAddrR(uint8_t sla)
{
	uint8_t	isOK = TWI_WriteByte((sla << 1) | SLA_R_BIT);
	if (!isOK) { errorState = USI_TWI_NO_ACK_ON_ADDRESS; }
	return isOK;
}

//データを送信する
//戻り値	非0:成功, 0:失敗
uint8_t	TWI_WriteByte(uint8_t data)
{
	uint8_t	retData;
	errorState = USI_TWI_SUCCESS;

	/* Write a byte */
	USIDR = data;							// Setup data.
	MasterTransmit( BV_USISR_8BITSHIFT );	// Send 8 bits on bus.

	/* Clock and verify (N)ACK from slave */
	//・受信データのLSB側1ビットがACK(0)/NACK(1)を示している。
	retData = MasterReceive( BV_USISR_1BITSHIFT );
	if (retData & 0x01) { errorState = USI_TWI_NO_ACK_ON_DATA; }

	return (errorState == USI_TWI_SUCCESS);
}

//データを受信する
//引数	isLastOne:	非0:この読み込みが最後(NACKを返す), 0:この後にまだ読み込む(ACKを返す)
//		data:	読み込んだ値が格納される
//戻り値	非0:成功, 0:失敗
//
//【例：I2C-EEPROMのシーケンシャル読み込み】
//・ ACKを返す… データを1個受信し、続けて次のデータを要求する場合。
//・NACKを返す… データを1個受信して完了する場合（次にストップを送信する）。
uint8_t	TWI_ReadByte(uint8_t isLastOne, uint8_t* data)
{
	errorState = USI_TWI_SUCCESS;	//特にエラー判定はないが。

	/* Read a data byte */
	*data = MasterReceive( BV_USISR_8BITSHIFT );

	/* Prepare to generate ACK (or NACK in case of End Of Transmission) */
	//	If transmission of last byte was performed.
	//	Load NACK to confirm End Of Transmission. (0xFF)
	//	Load ACK. Set data register bit 7 (output for SDA) low. (0x00)
	USIDR = isLastOne ? 0xFF : 0x00;
	
	MasterTransmit( BV_USISR_1BITSHIFT );
	
	return (errorState == USI_TWI_SUCCESS);
}

//ストップを送信する
//・SCLがHiの状態でSDAをLo→Hiにする。
//・データ送信後、SCLはLo、SDAはHi(NACK)かLo(ACK)。まずSDAをLoにする。
//・ストップ送信後はSCL,SDAともHiの状態（バスを解放した状態）。
uint8_t	TWI_Stop()
{
	errorState = USI_TWI_SUCCESS;

	USIDR = 0xFF;						//"released level" data. (必要)
	PORT_USI &= ~(1<<PIN_USI_SDA);		// Pull SDA low.

	//ストップコンディションを生成
	PORT_USI |= (1<<PIN_USI_SCL);		// Release SCL.
	_delay_us(T_SETUP_STOP);
	PORT_USI |= (1<<PIN_USI_SDA);		// Release SDA.
	_delay_us(T_BUS_FREE);

#ifdef SIGNAL_VERIFY
	if( !(USISR & (1<<USIPF)) ) { errorState = USI_TWI_MISSING_STOP_CON; }
#endif

	return (errorState == USI_TWI_SUCCESS);
}
